from napalm_srl import NokiaSRLDriver
from napalm_srl.srl import SRLAPI
from napalm.base import models as napalm_models
from ..base import SCBaseNetconfDriver

from ipaddress import IPv6Address


class SCSLRAPI(SRLAPI):
    """
    This override fixes an issue with IPv6 addresses embedded in URLs when talking
    over GRPCs. Not sure if it's moot with the various TLS
    """

    def __init__(self, hostname, username, password, timeout=60, optional_args=None):
        super().__init__(hostname, username, password, timeout, optional_args)

        try:
            IPv6Address(hostname)
            self.target = f"[{hostname}]:{self.gnmi_port}"
        except ValueError:
            pass


class SCNokiaSRLDriver(NokiaSRLDriver, SCBaseNetconfDriver):
    def __init__(self, hostname, username, password, timeout=60, optional_args=None):
        """
        Forcing insecure connection for testing purposes
        """
        optional_args = optional_args if optional_args else {}
        optional_args["insecure"] = True
        optional_args["skip_verify"] = True

        optional_args["encoding"] = "JSON_IETF"

        super().__init__(
            hostname, username, password, timeout=timeout, optional_args=optional_args
        )

        self.device = SCSLRAPI(
            hostname, username, password, timeout=timeout, optional_args=optional_args
        )

    def get_config(
        self, retrieve="all", full=False, sanitized=False, format="text"
    ) -> napalm_models.ConfigDict:
        """
        pulling running config via ssh sow we can have it in 'display set' format.
        """
        if format != "text":
            return super().get_config(
                retrieve=retrieve, full=full, sanitized=sanitized, format=format
            )

        config = {"startup": "", "running": "", "candidate": ""}

        with self.ssh_conn() as ssh_device:
            if retrieve in ["all", "running"]:
                config["running"] = ssh_device.send_command("info from running flat")
            if retrieve in ["all", "startup"]:
                config["startup"] = ssh_device.send_command("info from startup flat")
            if retrieve in ["all", "candidate"]:
                config["candidate"] = ssh_device.send_command(
                    "info from candidate flat"
                )

        return config

    def get_optics(self):
        path = {"/interface"}
        path_type = "STATE"
        output = self.device._gnmiGet("", path, path_type)
        interfaces = self._getObj(
            output, *["srl_nokia-interfaces:interface"], default=[]
        )
        channel_data = {}

        for i in interfaces:
            if not self._getObj(i, *["transceiver", "channel"], default=False):
                continue

            name = self._getObj(i, *["name"])
            channel = self._getObj(i, *["transceiver", "channel"], default={})[0]
            channel_data.update(
                {
                    name: {
                        "physical_channels": {
                            "channel": [
                                {
                                    "index": self._getObj(
                                        channel, *["index"], default=-1
                                    ),
                                    "state": {
                                        "input_power": {
                                            "instant": self._getObj(
                                                channel,
                                                *["input-power", "latest-value"],
                                                default=-1.0,
                                            ),
                                            "avg": -1.0,
                                            "min": -1.0,
                                            "max": -1.0,
                                        },
                                        "output_power": {
                                            "instant": self._getObj(
                                                channel,
                                                *["output-power", "latest-value"],
                                                default=-1.0,
                                            ),
                                            "avg": -1.0,
                                            "min": -1.0,
                                            "max": -1.0,
                                        },
                                        "laser_bias_current": {
                                            "instant": self._getObj(
                                                channel,
                                                *["laser-bias-current", "latest-value"],
                                                default=-1.0,
                                            ),
                                            "avg": -1.0,
                                            "min": -1.0,
                                            "max": -1.0,
                                        },
                                    },
                                }
                            ]
                        }
                    }
                }
            )

        return channel_data

    def get_inventory(self):
        path = {"/interface/transceiver"}
        path_type = "STATE"
        output = self.device._gnmiGet("", path, path_type)
        transceivers = self._getObj(
            output, *["srl_nokia-interfaces:interface"], default=[]
        )
        result = []
        for t_if in transceivers:
            trans = t_if["transceiver"]

            if "serial-number" not in trans:
                continue

            result.append(
                {
                    "type": "optic",
                    "subtype": trans["ethernet-pmd"],
                    "name": t_if["name"],
                    "part_number": trans["vendor-part-number"],
                    "serial_number": trans["serial-number"],
                }
            )

        return result



from typing import Dict, List

from ncclient import manager
from ncclient.xml_ import to_ele
from lxml import etree

from napalm.base import NetworkDriver
from napalm.base.exceptions import ConnectionException
from napalm.base import models as napalm_models


import logging

from ..base import SCBaseNetconfDriver
from .. import models as sc_models
from . import constants as C
from .constants import NS

logger = logging.getLogger(__name__)


class SCNOSDriver(NetworkDriver, SCBaseNetconfDriver):
    """
    Drivenets class is heavily based on Napalm IOSXR, which is
    also a netconf/yang device.
    """

    netmiko_host_type = "linux"

    def __init__(self, hostname, username, password, timeout=60, optional_args=None):
        self.hostname = hostname
        self.username = username
        self.password = password
        self.timeout = timeout
        self.pending_changes = False
        self.replace = False
        self.locked = False
        self.optional_args = optional_args if optional_args else {}
        self.port = self.optional_args.pop("port", 830)
        self.key_file = self.optional_args.pop("key_file", None)
        if "ssh_config_file" in self.optional_args:
            #    self.optional_args["ssh_config"] = self.optional_args["ssh_config_file"]
            del self.optional_args["ssh_config_file"]

        self.optional_args["hostkey_verify"] = False
        self.device = None
        self.NS = NS

    def open(self):
        """Open the connection with the device."""
        try:
            self.device = manager.connect(
                host=self.hostname,
                port=self.port,
                username=self.username,
                password=self.password,
                key_filename=self.key_file,
                timeout=self.timeout,
                **self.optional_args,
            )

        except Exception as conn_err:
            logger.error(conn_err.args[0])
            raise ConnectionException(conn_err.args[0])

    def close(self):
        """Close the connection."""
        logger.debug("Closed connection with device %s" % (self.hostname))
        self.device.close_session()

    def get_config(
        self, retrieve="all", full=False, sanitized=False, format="text"
    ) -> napalm_models.ConfigDict:
        """
        Get config adapted from iosxr_netconf driver. Note that there doesn't appear to
        be a way to get a non-xml version of the config via netconf, so we're doing
        in a separate SSH CLI session.
        """

        if sanitized:
            raise NotImplementedError("Sanitized nos config not supported")
        if format not in ["text", "xml"]:
            raise NotImplementedError("Only XML and text config formats supported")

        config = {"startup": "", "running": "", "candidate": ""}

        # for text format we're going to SSH into the device and do 'show config'
        # in this scenario 'candidate' isn't really relevant.
        if format == "text":
            with self.ssh_conn() as ssh_device:
                config["running"] = ssh_device.send_command("show config")

        # otherwise we're just doing netconf 'get-config'
        else:
            if retrieve.lower() in ["running", "all"]:
                config["running"] = str(self.device.get_config(source="running").xml)
            if retrieve.lower() in ["candidate", "all"]:
                config["candidate"] = str(
                    self.device.get_config(source="candidate").xml
                )

            for datastore in config:
                if not config[datastore]:
                    continue

                config[datastore] = etree.tostring(
                    etree.fromstring(config[datastore][0]),
                    pretty_print=True,
                    encoding="unicode",
                )

        return config

    def get_facts(self) -> napalm_models.FactsDict:
        """napalm get_facts"""
        rpc_reply = self.device.dispatch(to_ele(C.FACTS_RPC_REQ)).xml
        facts_xml = etree.fromstring(rpc_reply)

        return {
            "vendor": "Drivenets",
            "os_version": self._text(facts_xml, "//dn-sys:system-version", default=""),
            "hostname": self._text(facts_xml, "//dn-sys:name", default=""),
            "uptime": self._text(facts_xml, "//dn-sys:system-uptime", default=-1.0),
            "serial_number": self._text(
                facts_xml, "//dn-platform:serial-number", default=""
            ),
            "fqdn": self._text(facts_xml, "//dn-sys-dns:domain-name", default=""),
            "model": self._text(facts_xml, "//dn-sys:system-type", default=""),
            "interface_list": [i.text for i in self._xpath(facts_xml, "//dn-if:name")],
        }

    def get_optics(self) -> Dict[str, napalm_models.OpticsDict]:
        """napalm get_optics"""
        rpc_reply = self.device.dispatch(to_ele(C.OPTICS_RPC_REQ)).xml
        optics_xml = etree.fromstring(rpc_reply)

        # print(etree.tostring(optics_xml, pretty_print=True, encoding="unicode"))

        output = {}
        for i in self._xpath(optics_xml, "//dn-if:interface"):
            if not self._text(i, ".//dn-trans:transceiver-voltage"):
                continue
            i_name = self._text(i, "dn-if:name")
            output[i_name] = {"physical_channels": {"channels": {}}}

            i_channels = output[i_name]["physical_channels"]["channels"]
            for c in self._xpath(i, ".//dn-trans:channel"):
                if self._text(c, ".//dn-trans:receive-power"):
                    i_channels[self._text(c, ".//dn-trans:lane")] = {
                        "input_power": self._text(c, ".//dn-trans:receive-power"),
                        "output_power": self._text(c, ".//dn-trans:transmit-power"),
                        "laser_bias_current": self._text(
                            c, ".//dn-trans:laser-bias-current"
                        ),
                    }

        return output

    def get_lldp_neighbors(self) -> Dict[str, List[napalm_models.LLDPNeighborDict]]:
        """napalm get lldp neighbors"""
        rpc_reply = self.device.dispatch(to_ele(C.LLDP_NEIGH_RPC_REQ)).xml
        lldp_xml = etree.fromstring(rpc_reply)

        # print(etree.tostring(lldp_xml, pretty_print=True, encoding="unicode"))

        output = {}
        for i in self._xpath(lldp_xml, "//dn-lldp:interface"):
            if not self._xpath(i, ".//dn-lldp:system-name"):
                continue
            i_name = self._text(i, ".//dn-lldp:name")
            output[i_name] = []

            for n in self._xpath(i, ".//dn-lldp:neighbor"):
                output[i_name].append(
                    {
                        "hostname": self._text(n, ".//dn-lldp:system-name"),
                        "port": self._text(n, ".//dn-lldp:port-id"),
                    }
                )

        return output

    def get_inventory(self) -> List[sc_models.InventoryDict]:
        """
        sc-napalm get inventory
        """
        rpc_reply = self.device.dispatch(to_ele(C.INVENTORY_RPC_REQ)).xml
        inv_xml = etree.fromstring(rpc_reply)
        # print(etree.tostring(inv_xml, pretty_print=True, encoding="unicode"))

        output = []
        for i in self._xpath(inv_xml, "//dn-if:interface"):
            if not self._xpath(i, ".//dn-trans:ethernet-pmd"):
                continue

            i_name = self._text(i, "//dn-if:name")
            output.append(
                {
                    "type": "optic",
                    "subtype": self._get_optic_subtype(i),
                    "name": i_name,
                    "part_number": self._text(i, ".//dn-trans:vendor-part").strip(),
                    "serial_number": self._text(i, ".//dn-trans:serial-no").strip(),
                }
            )

        return output

    def _get_optic_subtype(self, i):
        """
        Optic subtypes are a bit tricky and sometimes have to be
        inferred
        """
        # the ethernet-pmd attribute gives us an official
        # 'transport type' which matches the data we want here
        eth_pmd = self._text(i, ".//dn-trans:ethernet-pmd")
        eth_pmd = eth_pmd.replace("dn-transport-types:", "")

        # terrible assumptions - right now on the one nos switch I only
        # see 'undefined' for 400G and 100G, wavelengths are all "1311"
        if eth_pmd == "ETH_UNDEFINED":
            speed = int(self._text(i, ".//dn-if:interface-speed")) / 1000
            if self._text(i, ".//dn-trans:wavelength") == "1311":
                return f"ETH_{int(speed)}GBASE-LR4"

            raise ValueError(
                f"ETH_UNDEFINED {i[0].text} speed {speed} WAVELENGTH {self._text(i, './/dn-trans:wavelength')}"
            )

        return eth_pmd



from pathlib import Path
import os
import json
from inspect import isclass
import re

from google.protobuf import json_format
import custom_napalm
from jnpr.junos.factory.optable import OpTable

MOCK_DATA_DIR = Path(__file__).parent / "mock_data"


def mock_netmiko_send(path: str, cmd: str, encoding="text") -> str:
    """
    Generic 'translate a device command to a valid filename, then read
    data from that file'
    path: Path object pointing to where that file resides
    cmd: the command string passed to the relevant netmiko/junper/etc "send_command" method
    """
    cmd = cmd.replace(" ", "_")
    cmd = cmd.replace("*", "star")

    if encoding == "json":
        cmd += ".json"
    else:
        cmd += ".txt"

    cmd_file = str(path + "/" + cmd)

    with open(cmd_file, encoding="utf-8") as fh:
        cmd_txt = fh.read()

    return cmd_txt


class MockDriver:
    def __init__(self, test_dir: Path, **kwargs):
        self.test_dir = test_dir

    def open(self):
        pass

    def close(self):
        pass


class MockNCManager:
    """
    Fake ncclient 'manager' class that overrides 'dispatch'
    to pull xml data from a file and save it to an .xml
    attribute
    """

    def __init__(self, file_path: Path):
        self.file_path = file_path

    def dispatch(self, element):
        # we're going to assume that the RPC call is a `get` with a `filter`
        # and that the filename in the mock data corresponds to the highest
        # element in the filter.
        filter_no_ns = re.sub(r"\{.+\}", "", element[0][0].tag)
        file_name = str(self.file_path / f"get-{filter_no_ns}.xml")

        with open(file_name) as fh:
            data = fh.read()

        self.xml = data

        return self


class MockSRLAPI(object):
    def __init__(self, file_path: Path):
        self.file_path = file_path

    def open(self):
        pass

    def close(self):
        pass

    def _gnmiGet(self, prefix, paths, pathType):
        """
        for each path in the set of paths
        look for a file named based on that path with "/"
        replaced with "_".
        Contatenate the results of those files together
        """
        output = ""
        for path in paths:
            # removing leading slashes and indexes
            path = re.sub(r"^/", "", path)
            path = re.sub(r"\[.+\]", "", path)
            path = re.sub(r"[/-]", "_", path)
            cmd_file = f"{self.file_path}/{path}.txt"
            with open(cmd_file, encoding="utf-8") as fh:
                output += fh.read()
        print(output)

        output = re.sub(r"\'", '"', output)
        return json_format.ParseFromString(output)


class MockEOS(MockDriver, custom_napalm.SCEOSDriver):
    driver_name = "eos"

    def __init__(self, test_dir, **kwargs):
        MockDriver.__init__(self, test_dir, **kwargs)
        self.transport = "ssh"

    def _run_commands(self, cmds: list[str], encoding="json"):
        """
        Mocks eos _run_commands, assumes your file is json-encoded
        and returns a data structure based on that.
        """
        result = []
        for cmd in cmds:
            cmd_text = mock_netmiko_send(str(self.test_dir), cmd, encoding=encoding)
            try:
                cmd_json = json.loads(cmd_text)
            except json.decoder.JSONDecodeError:
                cmd_json = {}

            result.append(cmd_json)

        return result


class MockIOSXR(MockDriver, custom_napalm.SCIOSXR):
    driver_name = "iosxr"

    def open(self):
        self.device = MockNCManager(self.test_dir)


class MockJunos(MockDriver, custom_napalm.SCJunOS):
    driver_name = "junos"

    def __init__(self, test_dir, **kwargs):
        """
        With junos we are going to override the pyEZ initializer for
        """
        MockDriver.__init__(self, test_dir, **kwargs)
        self.device = None

        for obj in custom_napalm.junos.junos_views.__dict__.values():
            if isclass(obj) and issubclass(obj, OpTable):
                # only override initializer if we have mock data for this view
                mock_rpc = f"{self.test_dir}/{obj.GET_RPC}.xml"
                if not os.path.isfile(mock_rpc):
                    continue

                def new_init(obj, dev=None, xml=None, path=None, use_filter=True):
                    obj._dev = None
                    obj.view = obj.VIEW
                    obj._key_list = []
                    obj._path = f"{self.test_dir}/{obj.GET_RPC}.xml"
                    # obj._lxml = etree.parse(f"{self.path}/{obj.GET_RPC}.xml")
                    # obj.xml = etree.parse(f"{self.path}/{obj.GET_RPC}.xml")
                    obj._use_filter = obj.USE_FILTER and use_filter
                    if obj._dev is not None:
                        obj._use_filter = obj._use_filter and obj._dev._use_filter

                obj.__init__ = new_init


class MockNXOS(MockDriver, custom_napalm.SCNXOS):
    driver_name = "nxos"

    def _send_command(self, cmd: str):
        return mock_netmiko_send(str(self.test_dir), cmd)


class MockNokiaSRL(MockDriver, custom_napalm.SCNokiaSRLDriver):
    driver_name = "srl"

    def __init__(self, test_dir, **kwargs):
        MockDriver.__init__(self, test_dir, **kwargs)
        self.device = MockSRLAPI(test_dir)



from typing import Literal
from typing_extensions import TypedDict

VALID_INVENTORY_TYPES = Literal[
    "fabric_module",
    "fan",
    "linecard",
    "optic",
    "psu",
    "re",
    "stack_cable",
    "stack_member",
    "uplink_module",
    "aoc",
    "dac",
]

InventoryDict = TypedDict(
    "InventoryDict",
    {
        "type": VALID_INVENTORY_TYPES,
        "subtype": str,
        "name": str,
        "part_number": str,
        "serial_number": str,
    },
)
